# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple frontend to the Timesketch data importer."""
from __future__ import unicode_literals

import argparse
import getpass
import logging
import os
import sys

from typing import Dict

from timesketch_api_client import cli_input
from timesketch_api_client import credentials as ts_credentials
from timesketch_api_client import crypto
from timesketch_api_client import config
from timesketch_api_client import sketch
from timesketch_api_client import version as api_version
from timesketch_import_client import helper
from timesketch_import_client import importer
from timesketch_import_client import version as importer_version


logger = logging.getLogger('timesketch_importer.importer_frontend')


def configure_logger_debug():
    """Configure the logger to log debug logs."""
    logger.setLevel(logging.DEBUG)
    logger_formatter = logging.Formatter(
        '[%(asctime)s] %(name)s/%(levelname)s %(message)s '
        '<%(module)s/%(funcName)s>')
    for handler in logger.parent.handlers:
        handler.setFormatter(logger_formatter)


def configure_logger_default():
    """Configure the logger to only log information and above logs."""
    logger.setLevel(logging.INFO)
    logger_formatter = logging.Formatter(
        '[%(asctime)s] %(name)s/%(levelname)s %(message)s')
    for handler in logger.parent.handlers:
        handler.setFormatter(logger_formatter)


def upload_file(
        my_sketch: sketch.Sketch, config_dict: Dict[str, any],
        file_path: str) -> str:
    """Uploads a file to Timesketch.

    Args:
        my_sketch (sketch.Sketch): a sketch object to point to the sketch the
            data will be imported to.
        config_dict (dict): dict with settings for the importer.
        file_path (str): the path to the file to upload.

    Returns:
        A string with results (whether successful or not).
    """
    if not my_sketch or not hasattr(my_sketch, 'id'):
        return 'Sketch needs to be set'

    _, _, file_extension = file_path.rpartition('.')
    if file_extension.lower() not in ('plaso', 'csv', 'jsonl'):
        return (
            'File needs to have one of the following extensions: '
            '.plaso, .csv, '
            '.jsonl (not {0:s})').format(file_extension.lower())

    import_helper = helper.ImportHelper()
    import_helper.add_config_dict(config_dict)

    log_config_file = config_dict.get('log_config_file', '')
    if log_config_file:
        import_helper.add_config(log_config_file)

    with importer.ImportStreamer() as streamer:
        streamer.set_sketch(my_sketch)
        streamer.set_config_helper(import_helper)

        format_string = config_dict.get('message_format_string')
        if format_string:
            streamer.set_message_format_string(format_string)

        timeline_name = config_dict.get('timeline_name')
        if timeline_name:
            streamer.set_timeline_name(timeline_name)

        index_name = config_dict.get('index_name')
        if index_name:
            streamer.set_index_name(index_name)

        time_desc = config_dict.get('timestamp_description')
        if time_desc:
            streamer.set_timestamp_description(time_desc)

        entry_threshold = config_dict.get('entry_threshold')
        if entry_threshold:
            streamer.set_entry_threshold(entry_threshold)

        size_threshold = config_dict.get('size_threshold')
        if size_threshold:
            streamer.set_filesize_threshold(size_threshold)

        streamer.add_file(file_path)

    return 'File got successfully uploaded to sketch: {0:d}'.format(
        my_sketch.id)


def main(args=None):
    """The main function of the tool."""
    if args is None:
        args = sys.argv[1:]

    argument_parser = argparse.ArgumentParser(
        description='A tool to upload data to Timesketch, using the API.')

    argument_parser.add_argument(
        '--version', action='store_true', dest='show_version',
        help='Print version information')

    argument_parser.add_argument(
        '--debug', '--verbose', '-d', action='store_true', dest='show_debug',
        help='Make the logging more verbose to include debug logs.')

    auth_group = argument_parser.add_argument_group('Authentication Arguments')
    auth_group.add_argument(
        '-u', '--user', '--username', action='store', dest='username',
        type=str, help='The username of the Timesketch user.')
    auth_group.add_argument(
        '-p', '--password', '--pwd', action='store', type=str, dest='password',
        help=(
            'If authenticated with password, provide the password on the CLI. '
            'If neither password is provided nor a password prompt an OAUTH '
            'connection is assumed.'))
    auth_group.add_argument(
        '--pwd-prompt', '--pwd_prompt', action='store_true', default=False,
        dest='pwd_prompt', help='Prompt for password.')
    auth_group.add_argument(
        '--cred-prompt', '--cred_prompt', '--token-password',
        '--token_password', '--token', action='store_true', default=False,
        dest='cred_prompt',
        help='Prompt for password to decrypt and encrypt credential file.')
    auth_group.add_argument(
        '--client-secret', '--client_secret', action='store', type=str,
        default='', dest='client_secret', help='OAUTH client secret.')
    auth_group.add_argument(
        '--client-id', '--client_id', action='store', type=str, default='',
        dest='client_id', help='OAUTH client ID.')
    auth_group.add_argument(
        '--run_local', '--run-local', action='store_true', dest='run_local',
        help=(
            'If OAUTH is used to authenticate and the connection is over '
            'SSH then it is recommended to set this option. When set an '
            'authentication URL is prompted on the screen, requiring a '
            'copy/paste into a browser to complete the OAUTH dance.'))

    config_group = argument_parser.add_argument_group(
        'Configuration Arguments')
    config_group.add_argument(
        '--log-config-file', '--log_config_file', '--lc', action='store',
        type=str, default='', metavar='FILEPATH', dest='log_config_file',
        help=(
            'Path to a YAML config file that defines the config for parsing '
            'and setting up file parsing. By default formatter.yaml that '
            'comes with the importer will be used.'))
    config_group.add_argument(
        '--host', '--hostname', '--host-uri', '--host_uri', dest='host_uri',
        type=str, default='', action='store',
        help='The URI to the Timesketch instance')
    config_group.add_argument(
        '--format_string', '--format-string', type=str, action='store',
        dest='format_string', default='', help=(
            'Formatting string for the message field. If there is no message '
            'field in the input data a message string can be composed using '
            'a format string.'))
    config_group.add_argument(
        '--timeline_name', '--timeline-name', action='store', type=str,
        dest='timeline_name', default='', help=(
            'String that will be used as the timeline name.'))
    config_group.add_argument(
        '--index-name', '--index_name', action='store', type=str, default='',
        dest='index_name', help=(
            'If the data should be imported into a specific timeline the '
            'index name needs to be provided, otherwise a new index will '
            'be generated.'))
    config_group.add_argument(
        '--timestamp_description', '--timestamp-description', '--time-desc',
        '--time_desc', action='store', type=str, default='', dest='time_desc',
        help='Value for the timestamp_description field.')
    config_group.add_argument(
        '--threshold_entry', '--threshold-entry', '--entries', action='store',
        type=int, default=0, dest='entry_threshold',
        help=(
            'How many entries should be buffered up before being '
            'sent to server.'))
    config_group.add_argument(
        '--threshold_size', '--threshold-size', '--filesize', action='store',
        type=int, default=0, dest='size_threshold',
        help=(
            'For binary file transfer, how many bytes should be transferred '
            'per chunk.'))
    config_group.add_argument(
        '--sketch_id', '--sketch-id', type=int, default=0, dest='sketch_id',
        action='store', help=(
            'The sketch ID to store the timeline in, if no sketch ID is '
            'provided a new sketch will be created.'))

    argument_parser.add_argument(
        'path', action='store', nargs='?', type=str, help=(
            'Path to the file that is to be imported.'))

    options = argument_parser.parse_args(args)

    if options.show_version:
        print('API Client Version: {0:s}'.format(api_version.get_version()))
        print('Importer Client Version: {0:s}'.format(
            importer_version.get_version()))
        sys.exit(0)

    if options.show_debug:
        configure_logger_debug()
    else:
        configure_logger_default()

    if not options.path:
        logger.error(
            'A valid file path needs to be provided, unable to continue.')
        sys.exit(1)

    if not os.path.isfile(options.path):
        logger.error('Path {0:s} is not valid, unable to continue.'.format(
            options.path))
        sys.exit(1)

    assistant = config.ConfigAssistant()
    assistant.load_config_file()
    assistant.load_config_dict(vars(options))

    cred_storage = crypto.CredentialStorage()
    token_password = ''
    if options.cred_prompt:
        token_password = cli_input.ask_question(
            'Enter password to encrypt/decrypt credential file',
            input_type=str, hide_input=True)

    try:
        credentials = cred_storage.load_credentials(
            config_assistant=assistant, password=token_password)
    except IOError:
        logger.error(
            'Unable to decrypt the credential file')
        logger.debug(
            'Error details:', exc_info=True)
        logger.error(
            'If you\'ve forgotten the password you can delete '
            'the ~/.timesketch.token file and run the tool again.')
        sys.exit(1)

    conf_password = ''

    if credentials:
        if credentials.TYPE.lower() == 'oauth':
            assistant.set_config('auth_mode', 'oauth')
        elif credentials.TYPE.lower() == 'timesketch':
            assistant.set_config('auth_mode', 'timesketch')

    else:
        # Check whether we'll need to read password for timesketch
        # auth from the command line.
        if options.pwd_prompt:
            conf_password = getpass.getpass('Type in the password: ')
        else:
            conf_password = options.password

        if options.run_local:
            assistant.set_config('auth_mode', 'oauth_local')

        if options.client_secret:
            assistant.set_config('auth_mode', 'oauth')

        if conf_password:
            assistant.set_config('auth_mode', 'timesketch')

    if conf_password:
        credentials = ts_credentials.TimesketchPwdCredentials()
        credentials.credential = {
            'username': assistant.get_config('username'),
            'password': conf_password
        }
        logger.info('Saving Credentials.')
        cred_storage.save_credentials(
            credentials, password=token_password,
            config_assistant=assistant)

    # Gather all questions that are missing.
    config.configure_missing_parameters(
        config_assistant=assistant, token_password=token_password)

    logger.info('Creating a client.')
    ts_client = assistant.get_client(token_password=token_password)
    if not ts_client:
        logger.error('Unable to create a Timesketch API client, exiting.')
        sys.exit(1)

    logger.info('Client created.')
    logger.info('Saving TS config.')
    assistant.save_config()

    if ts_client.credentials:
        logger.info('Saving Credentials.')
        cred_storage.save_credentials(
            ts_client.credentials, password=token_password,
            config_assistant=assistant)

    sketch_id = options.sketch_id
    if sketch_id:
        my_sketch = ts_client.get_sketch(sketch_id)
    else:
        my_sketch = ts_client.create_sketch('New Sketch From Importer CLI')

    if not my_sketch:
        logger.error('Unable to get sketch ID: {0:d}'.format(sketch_id))
        sys.exit(1)

    filename = os.path.basename(options.path)
    default_timeline_name, _, _ = filename.rpartition('.')

    if options.timeline_name:
        conf_timeline_name = options.timeline_name
    else:
        conf_timeline_name = cli_input.ask_question(
            'What is the timeline name', input_type=str,
            default=default_timeline_name)

    config_dict = {
        'message_format_string': options.format_string,
        'timeline_name': conf_timeline_name,
        'index_name': options.index_name,
        'timestamp_description': options.time_desc,
        'entry_threshold': options.entry_threshold,
        'size_threshold': options.size_threshold,
        'log_config_file': options.log_config_file,
    }

    logger.info('Uploading file.')
    result = upload_file(
        my_sketch=my_sketch, config_dict=config_dict, file_path=options.path)
    logger.info(result)


if __name__ == '__main__':
    main()



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Some utility functions for the Timesketch importer."""

import datetime
import string

import six
import dateutil.parser


# When there is no defined format message string a default one is used
# that is generated using the available columns or keys in the dataset.
# This constant defines what columns or keys should be skipped while
# generating this message field.
FIELDS_TO_SKIP_IN_FORMAT_STRING = frozenset([
    'timestamp_desc', 'time', 'timestamp', 'data_type', 'datetime',
    'source_short', 'source_long', 'source'
])
# This constant defines patterns that are used to skip columns or keys
# that start with the strings defined here.
FIELDS_TO_SKIP_STARTSWITH = frozenset([
    '_',
])


def format_data_frame(dataframe, format_message_string):
    """Add a message field to a data frame using a format message string.

    Args:
        dataframe (pandas.DataFrame): the data frame containing the data.
        format_message_String (str): the format string used to generate
            a message column.
    """
    dataframe['message'] = ''

    formatter = string.Formatter()
    for literal_text, field, _, _ in formatter.parse(format_message_string):
        dataframe['message'] = dataframe['message'] + literal_text

        if field:
            dataframe['message'] = dataframe[
                'message'] + dataframe[field].astype(str)


def get_combined_message_string(dataframe=None, mydict=None):
    """Returns a message string by combining all columns from a DataFrame/dict.

    Args:
        dataframe (pandas.DataFrame): the data frame containing the data. Used
            to get column names for the message string if supplied.
        mydict (dict): if supplied the dictionary keys will be used to
            construct the message string.

    Raises:
        ValueError if neither dataframe nor a dictionary is supplied.

    Returns:
        A string that can be used as a default message string.
    """
    if dataframe is None and mydict is None:
        raise ValueError('Need to define either a dict or a DataFrame')

    if mydict:
        my_list = list(mydict.keys())
    else:
        my_list = list(dataframe.columns)

    fields_to_delete = list(set(my_list).intersection(
        FIELDS_TO_SKIP_IN_FORMAT_STRING))
    for field in fields_to_delete:
        index = my_list.index(field)
        _ = my_list.pop(index)

    for del_field in FIELDS_TO_SKIP_STARTSWITH:
        for index, field in enumerate(my_list):
            if field.startswith(del_field):
                _ = my_list.pop(index)

    string_values = [
        '[{0:s}] = {{{0:s}}}'.format(x) for x in my_list]
    return ', '.join(string_values)


def get_datestring_from_value(value):
    """Returns an empty string or a date value.

    Args:
        value (any): A string or an int/float that contains a timestamp
            value that can be parsed into a datetime.

    Returns:
        A string with the timestamp in ISO 8601 or an empty string if not
        able to parse the timestamp.
    """
    if isinstance(value, six.string_types):
        try:
            date = dateutil.parser.parse(value)
            return date.isoformat()
        except ValueError:
            pass

    if isinstance(value, (int, float)):
        try:
            date = datetime.datetime.utcfromtimestamp(value / 1e6)
            return date.isoformat()
        except ValueError:
            pass
    return ''



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch data import configuration."""
from __future__ import unicode_literals

import codecs
import logging
import os

import yaml


logger = logging.getLogger('timesketch_importer.config_loader')

DEFAULT_FILE = 'formatter.yaml'


def load_config(file_path=''):
    """Loads YAML config and returns a list of dict with the results.

    Args:
        file_path (str): path to the YAML config file. This is optional
            and if not defined the default formatter.yaml file will be
            used that comes with the tool.

    Returns:
        dict: a dict with the key being a config file identifier and the value
        being another dict with the configuration items.
    """
    if not file_path:
        base_path = os.path.dirname(__file__)
        file_path = os.path.join(base_path, DEFAULT_FILE)

    if not file_path.endswith('.yaml'):
        logger.error('Can\'t load a config that is not a YAML file.')
        return {}

    if not os.path.isfile(file_path):
        logger.error('File path does not exist, unable to load YAML config.')
        return {}

    with codecs.open(file_path, 'r') as fh:
        try:
            data = yaml.safe_load(fh)
            return data
        except (AttributeError, yaml.parser.ParserError) as e:
            logger.error('Unable to parse YAML file, with error: %s', e)
            return {}

    return {}



#!/usr/bin/env python
# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This is the setup file for the project."""
from __future__ import unicode_literals

import os
import glob

from setuptools import find_packages
from setuptools import setup

from timesketch_import_client import version

setup(
    name='timesketch-import-client',
    version=version.get_version(),
    description='Timesketch Import Client',
    license='Apache License, Version 2.0',
    url='http://www.timesketch.org/',
    maintainer='Timesketch development team',
    maintainer_email='timesketch-dev@googlegroups.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ],
    data_files=[
        ('data', glob.glob(
            os.path.join('timesketch_import_client', 'data', '*.yaml'))),
    ],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'timesketch_importer = tools.timesketch_importer:main']},
    install_requires=frozenset([
        'pandas',
        'xlrd',
        'timesketch-api-client',
        'pyyaml',
    ]),
    )



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Timesketch data importer."""
from __future__ import unicode_literals

import logging
import os

from timesketch_import_client import data as data_config


logger = logging.getLogger('timesketch_importer.ts_import_helper')


class ImportHelper:
    """Import helper class."""

    def __init__(self, load_default=True):
        """Initialize the helper class."""
        if load_default:
            self._data = data_config.load_config()
        else:
            self._data = {}

    def _configure_streamer(self, streamer, config_dict):
        """Sets up the streamer based on a configuration dict.

        Args:
            streamer (ImportStreamer): an import streamer object.
            config_dict (dict): A dictionary that contains
                configuration details for the streamer.
        """
        message = config_dict.get('message')
        if message:
            streamer.set_message_format_string(message)

        timestamp_desc = config_dict.get('timestamp_desc')
        if timestamp_desc:
            streamer.set_timestamp_description(timestamp_desc)

        separator = config_dict.get('separator')
        if separator:
            streamer.set_csv_delimiter(separator)

        encoding = config_dict.get('encoding')
        if encoding:
            streamer.set_text_encoding(encoding)

        datetime_string = config_dict.get('datetime')
        if datetime_string:
            streamer.set_datetime_column(datetime_string)

    def add_config(self, file_path):
        """Loads a YAML config file describing the log file config.

        This function reads a YAML config file, and adds the config
        to it's config data. This configuration can then be used to
        setup an import streamer object.

        Args:
            file_path (str): the path to the config file.

        Raises:
            ValueError: if the file path does not exist or if the config
                        is not valid or cannot be read.
        """
        if not os.path.isfile(file_path):
            raise ValueError(
                'Unable to open file: [{0:s}], it does not exist.'.format(
                    file_path))

        if not os.access(file_path, os.R_OK):
            raise ValueError(
                'Unable to open file: [{0:s}], cannot open it for '
                'read, please check permissions.'.format(file_path))


        config = data_config.load_config(file_path)
        if not isinstance(config, dict):
            raise ValueError(
                'Unable to read config file since it does not produce a dict')

        if not all([isinstance(x, dict) for x in config.values()]):
            raise ValueError(
                'The config needs to a dict that contains other dict '
                'attributes.')

        self._data.update(config)

    def add_config_dict(self, config, config_name='manual'):
        """Add a config dict describing the log file config.

        Args:
            config (dict): a single dict with the config needed for setting
                up the streamer.
            config_name (str): the name of the config to be added. This is
                optional, with the default value set to "manual".
        """
        self._data[config_name] = config

    def configure_streamer(self, streamer, data_type='', columns=None):
        """Go through loaded config and setup a streamer if there is a match.

        This function takes a streamer object and compares the loaded config
        to see if there is a match to the data_type and/or columns that are
        supplied to the function and sets the streamer up if a matching
        config is discovered.

        The helper will use the first match in the config to setup the
        streamer (it will exit as soon as it finds a match).

        Args:
            streamer (ImportStreamer): an import streamer object.
            data_type (str): optional data type that is used for matching
                with the loaded config.
            columns (List[str]): optional list of strings with column names.
        """
        for config_name, config in self._data.items():
            conf_data_type = config.get('data_type')
            if data_type and conf_data_type and conf_data_type == data_type:
                logger.info('Using config %s for streamer.', config_name)
                self._configure_streamer(streamer, config)
                return

            if not columns:
                continue

            column_set = set(columns)
            column_string = config.get('columns', '')
            column_subset_string = config.get('columns_subset', '')
            if not any([column_string, column_subset_string]):
                continue

            conf_columns = set(column_string.split(','))
            if conf_columns and column_set == conf_columns:
                logger.info('Using config %s for streamer.', config_name)
                self._configure_streamer(streamer, config)
                return

            columns_subset = set(column_subset_string.split(','))
            if columns_subset and columns_subset.issubset(column_set):
                logger.info('Using config %s for streamer.', config_name)
                self._configure_streamer(streamer, config)
                return
